import os
from pathlib import Path
from typing import Iterator, Optional

from pyserato.encoders.base_encoder import BaseEncoder
from pyserato.model.crate import Crate
from pyserato.model.track import Track
from pyserato.util import serato_encode, serato_decode

DEFAULT_SERATO_FOLDER = Path(os.path.expanduser("~/Music/_Serato_"))


class Builder:

    def __init__(self, encoder: Optional[BaseEncoder] = None):
        self._encoder = encoder

    @staticmethod
    def _resolve_path(root: Crate) -> Iterator[tuple[Crate, str]]:
        """
        DFS through the crate tree returning a generator of paths with the current crate as the root.
        :return:
        """
        path = ""
        crates = [(root, path)]
        while crates:
            crate, path = crates.pop()
            path += f"{crate.name}%%"
            children = crate.children
            if children:
                for child in children:
                    crates.append((child, path))
            yield crate, path.rstrip("%%") + ".crate"

    @staticmethod
    def _parse_crate_names(filepath: Path) -> Iterator[str]:
        for name in str(filepath.name).split("%%"):
            yield name.replace(".crate", "")

    def _build_crate_filepath(self, crate: Crate, serato_folder: Path) -> Iterator[tuple[Crate, Path]]:
        subcrate_folder = serato_folder / "SubCrates"
        subcrate_folder.mkdir(exist_ok=True)
        for crate, paths in self._resolve_path(crate):
            yield crate, subcrate_folder / paths

    def parse_crates_from_root_path(self, subcrate_path: Path) -> dict[str, Crate]:
        crate_name_to_crate: dict[str, Crate] = {}
        for f in subcrate_path.iterdir():
            if not f.name.endswith("crate"):
                continue
            crate = self._build_crates_from_filepath(f)
            if crate.name in crate_name_to_crate:
                crate += crate_name_to_crate[crate.name]
                crate_name_to_crate[crate.name] = crate
            else:
                crate_name_to_crate[crate.name] = crate
        return crate_name_to_crate

    def _build_crates_from_filepath(self, filepath: Path) -> Crate:
        """
        Builds the crate tree from an existing file path.
        :param filepath:
        :return:
        """
        crate_names = list(self._parse_crate_names(filepath))
        child_crate = None
        crate = None
        for crate_name in reversed(crate_names):
            if child_crate is None:
                crate = Crate(crate_name)
                for file_path in self._parse_crate_tracks(filepath):
                    track = Track.from_path(file_path)
                    crate.add_track(track)
                child_crate = crate
            else:
                crate = Crate(crate_name, children=[child_crate])
                child_crate = crate
        assert crate, f"no crates parsed from {filepath}"
        return crate

    @staticmethod
    def _parse_crate_tracks(filepath: Path) -> Iterator[Path]:
        crate_content = filepath.read_bytes()
        while crate_content:
            otrk_idx = crate_content.find("otrk".encode("utf-8"))
            if otrk_idx < 0:
                break
            assert "otrk".encode("utf-8") == crate_content[otrk_idx: otrk_idx + len("otrk")]
            ptrk_idx = crate_content.find("ptrk".encode("utf-8"))
            # Below would be the lgnth of the otrk data if needed for further parsing.
            # otrk_section = crate_content[otrk_idx + len("otrk"): ptrk_idx]
            # len_data = int.from_bytes(otrk_section, "big") - 8
            ptrk_section = crate_content[ptrk_idx:]
            track_name_length = int.from_bytes(ptrk_section[4:8], "big")
            track_name_encoded = ptrk_section[8: 8 + track_name_length]
            file_path = serato_decode(track_name_encoded)
            if not file_path.startswith("/"):
                file_path = "/" + file_path

            yield Path(file_path)
            crate_content = crate_content[ptrk_idx + 8 + track_name_length:]

    def _construct(self, crate: Crate) -> bytes:
        """
        Constructs the crate in bytes ready to save to disk.
        Also writes any cues and meta info as tags to the track.
        """
        header = "vrsn".encode("latin1")
        # byteorder is redundant here for a size of 1 but is a required arg in Python 3.10-
        header += (0).to_bytes(1, byteorder="big")
        header += (0).to_bytes(1, byteorder="big")
        header += serato_encode("81.0")
        header += serato_encode("/Serato ScratchLive Crate")

        # sorting = "osrt".encode('latin1')
        # default_sorting_type = "track"
        # default_sorting_rev = (1 << 8)
        # sorting += (len(default_sorting_type) * 2 + 17).to_bytes(4, 'big')
        # sorting += "tvcn".encode('latin1')
        # sorting += (len(default_sorting_type) * 2).to_bytes(4, 'big')
        # sorting += default_sorting_type.encode('utf-16')
        # sorting += "brev".encode('latin1')
        # sorting += (default_sorting_rev).to_bytes(5, 'big')
        DEFAULT_COLUMNS = ["track", "artist", "album", "length"]
        column_section = bytes()
        for column in DEFAULT_COLUMNS:
            column_section += "ovct".encode()
            column_section += (len(column) * 2 + 18).to_bytes(4, "big")
            column_section += "tvcn".encode()
            column_section += (len(column) * 2).to_bytes(4, "big")
            column_section += serato_encode(column)
            column_section += "tvcw".encode()
            column_section += (2).to_bytes(4, "big")
            column_section += "0".encode()
            column_section += "0".encode()

        playlist_section = bytes()
        if crate.tracks:
            for track in crate.tracks:
                if self._encoder:
                    self._encoder.write(track)
                absolute_track_path = Path(track.path).resolve()

                otrk_size = (len(str(absolute_track_path)) * 2 + 8).to_bytes(4, "big")
                ptrk_size = (len(str(absolute_track_path)) * 2).to_bytes(4, "big")
                playlist_section += "otrk".encode("latin1")
                playlist_section += otrk_size
                playlist_section += "ptrk".encode("latin1")
                playlist_section += ptrk_size
                playlist_section += serato_encode(str(absolute_track_path))

        contents = header + column_section + playlist_section
        return contents

    def save(
        self,
        root: Crate,
        save_path: Path = DEFAULT_SERATO_FOLDER,
        overwrite: bool = False,
    ):
        for crate, filepath in self._build_crate_filepath(root, save_path):
            if filepath.exists() and overwrite is False:
                continue
            buffer = self._construct(crate)
            filepath.write_bytes(buffer)



import struct
from dataclasses import dataclass
from io import BytesIO
from typing import Optional

from pyserato.model.serato_color import SeratoColor
from pyserato.model.hot_cue_type import HotCueType


def write_null_terminated_string(s: str) -> bytearray:
    """Encode string as UTF-8 with a null terminator."""
    return bytearray(s.encode("utf-8") + b"\x00")


def concat_bytearrays(arrays: list[bytearray]) -> bytearray:
    result = bytearray()
    for arr in arrays:
        result.extend(arr)
    return result


def encode_element(name: str, data: bytes) -> bytearray:
    """Encode element with name, 4-byte big-endian length, and data."""
    name_bytes = write_null_terminated_string(name)
    length_bytes = struct.pack(">I", len(data))  # big-endian uint32
    return concat_bytearrays([name_bytes, bytearray(length_bytes), bytearray(data)])


@dataclass
class HotCue:
    name: str
    type: HotCueType
    start: Optional[int]
    index: int
    end: Optional[int] = None
    is_locked: Optional[bool] = False
    color: SeratoColor = SeratoColor.RED

    def __repr__(self):
        return "Start: {start}{end} | Index: {index} | Name: {name} | Color: {color}".format(
            name=self.name,
            index=str(f"{self.index}").rjust(2),
            start=str(f"{self.start}ms").ljust(10),
            end=str(f" | End: {self.end}ms").ljust(10) if self.end is not None else "",
            color=self.color.name,
        )

    # def apply_offset(self):
    #     if self.offset is None:
    #         return
    #     self._update_positions(int(self.offset.get_value()))

    # def _update_positions(self, offset_value: int):
    #     if self.start is not None:
    #         old_start = self.start
    #         self.start += offset_value
    #         if self.start < 0:
    #             self.start = old_start
    #             raise self._value_error("Start time", self.start, old_start)

    #     if self.end is not None:
    #         old_end = self.end
    #         self.end += offset_value
    #         if self.end < 0:
    #             self.end = old_end
    #             raise self._value_error("End time", self.end, old_end)

    # @staticmethod
    # def _value_error(offset_name: str, new_pos: int, old_pos: int) -> ValueError:
    #     return ValueError(f"{offset_name} cannot go below 0. New position: {new_pos}, old position: {old_pos}")

    def to_v2_bytes(self) -> bytes:
        if self.type == HotCueType.CUE:
            return self._cue_to_v2_bytes()
        elif self.type == HotCueType.LOOP:
            return self._loop_to_v2_bytes()
        raise ValueError(f"unsupported hotcue type {self.type}")

    def _loop_to_v2_bytes(self) -> bytes:
        """
                                            >B      c       I                   I               5s                      3s              >B      ?  # noqa: E501
        NAME    NULL    STRUCT LEN          NULL    INDEX   POS START           POS END         SOMETHING               COLOR                   LOCKED  NAME        NULL  # noqa: E501
        LOOP    \x00    \x00\x00\x00\x1f    \x00    \x00    \x00\x00\x00\xfe    \x00\x00\t%     \xff\xff\xff\xff\x00'   \xaa\xe1        \x00    \x00    first loop  \x00  # noqa: E501
        :param entry_data:
        :return:
        """
        name_bytes = write_null_terminated_string(self.name)
        buf = bytearray(0x14 + len(name_bytes))

        # flags / header fields
        buf[0x0] = 0x00
        buf[0x1] = self.index

        struct.pack_into(">I", buf, 0x02, self.start)  # big-endian uint32
        struct.pack_into(">I", buf, 0x06, self.end)  # big-endian uint32

        buf[0x0E] = 0
        buf[0x0F] = 0
        buf[0x10] = 255
        buf[0x11] = 255
        buf[0x13] = 1

        # append name string
        buf[0x14: 0x14 + len(name_bytes)] = name_bytes

        return bytes(encode_element("LOOP", buf))

    def _cue_to_v2_bytes(self) -> bytes:
        """
        NAME   NULL  STRUCT LEN          NULL    INDEX   POS START           POS END   COLOR  NULL  LOCKED  NAME        NULL  # noqa: E501
        CUE    \x00  \x00\x00\x00\x16    \x00    \x00    \x00\x00\x00\xfe    \x00      \xc0&& \x00  \x00    first bar   \x00  # noqa: E501
        """
        data = b"".join(
            (
                struct.pack(">B", 0),
                struct.pack(">B", self.index),
                struct.pack(">I", self.start),
                struct.pack(">B", 0),
                struct.pack(">3s", bytes.fromhex(self.color.value)),
                struct.pack(">B", 0),
                struct.pack(">?", b"\x00"),
                self.name.encode("utf-8"),
                struct.pack(">B", 0),
            )
        )
        return bytes(encode_element("CUE", data))

    @staticmethod
    def from_bytes(data: bytes, hotcue_type: HotCueType) -> "HotCue":
        fp = BytesIO(data)
        fp.seek(1)  # first byte is NULL as it's a separator
        if hotcue_type is HotCueType.CUE:
            result = (
                struct.unpack(">B", fp.read(1))[0],  # INDEX
                struct.unpack(">I", fp.read(4))[0],  # POSITION START
                struct.unpack(">B", fp.read(1))[0],  # NULL separator (aka POSITION END)
                None,  # aka. some field containing (4294967295, 39)
                struct.unpack(">3s", fp.read(3))[0],  # COLOR
                struct.unpack(">B", fp.read(1))[0],  # NULL separator
                struct.unpack(">?", fp.read(1))[0],  # LOCKED
                fp.read().partition(b"\x00")[0].decode("utf-8"),  # NAME + ending NULL separator
            )
            index, start, end, _1, color, _2, locked, name = result
            color = SeratoColor(color.hex().upper())
            hot_cue = HotCue(
                name=name,
                type=HotCueType.CUE,
                color=color,
                start=start,
                index=index,
            )
            return hot_cue
        elif hotcue_type is HotCueType.LOOP:
            index = struct.unpack(">B", fp.read(1))[0]
            start = struct.unpack(">I", fp.read(4))[0]
            end = struct.unpack(">I", fp.read(4))[0]

            # skip bytes 0x0E–0x13 (fixed flags and color placeholders)
            fp.seek(0x0E, 0)  # move to offset 0x0E in stream
            _ = fp.read(2)  # bytes 0x0E, 0x0F (color?)
            _ = fp.read(2)  # bytes 0x10, 0x11 (color?)
            _ = fp.read(1)  # byte 0x12 (padding 0)
            is_locked = struct.unpack(">B", fp.read(1))[0]  # byte 0x13 (locked = 1)

            # read the null-terminated name string
            name = fp.read().partition(b"\x00")[0].decode("utf-8")

            return HotCue(name=name, type=HotCueType.LOOP, start=start, end=end, index=index, is_locked=is_locked)
        else:
            raise ValueError(f"unknown type {hotcue_type}")



import struct
from io import BytesIO

from pyserato.model.hot_cue import HotCue
from pyserato.model.serato_color import SeratoColor
from pyserato.model.hot_cue_type import HotCueType


def test_to_v2_bytes_cue_starts_with_header():
    hc = HotCue(name="cue1", type=HotCueType.CUE, start=254, index=0, color=SeratoColor.RED)
    cue_bytes = hc.to_v2_bytes()
    assert cue_bytes.startswith(b"CUE\x00")


def test_to_v2_bytes_loop_starts_with_header():
    hc = HotCue(name="loop1", type=HotCueType.LOOP, start=254, end=2405, index=0)
    loop_bytes = hc.to_v2_bytes()
    assert loop_bytes.startswith(b"LOOP\x00")


def _get_entry_name(fp) -> str:
    entry_name = b""
    for x in iter(lambda: fp.read(1), b""):
        if x == b"\00":
            return entry_name.decode("utf-8")
        entry_name += x

    return ""


def test_cue_from_bytes_roundtrip():
    original = HotCue(name="cue_test", type=HotCueType.CUE, start=12345, index=3, color=SeratoColor.RED)
    bytes_data = original.to_v2_bytes()
    fp = BytesIO(bytes_data)
    entry_name = _get_entry_name(fp)
    assert entry_name == "CUE"
    struct_length = struct.unpack(">I", fp.read(4))[0]
    assert struct_length > 0  # normally this should not happen
    entry_data = fp.read(struct_length)
    parsed = HotCue.from_bytes(entry_data, HotCueType.CUE)
    assert parsed.name == original.name
    assert parsed.index == original.index
    assert parsed.start == original.start
    assert parsed.color == original.color


def test_loop_from_bytes_roundtrip():
    original = HotCue(name="loop_test", type=HotCueType.LOOP, start=12345, end=34567, index=3, color=SeratoColor.RED)
    bytes_data = original.to_v2_bytes()
    fp = BytesIO(bytes_data)
    entry_name = _get_entry_name(fp)
    assert entry_name == "LOOP"
    struct_length = struct.unpack(">I", fp.read(4))[0]
    assert struct_length > 0  # normally this should not happen
    entry_data = fp.read(struct_length)
    parsed = HotCue.from_bytes(entry_data, HotCueType.LOOP)
    assert parsed.name == original.name
    assert parsed.index == original.index
    assert parsed.start == original.start
    assert parsed.end == original.end
    assert parsed.color == original.color



from pathlib import Path
from pyserato.builder import DEFAULT_SERATO_FOLDER, Crate, Builder
from pyserato.model.track import Track


def list_crates(serato_folder: Path = DEFAULT_SERATO_FOLDER) -> list[Crate]:
    all_crates = []
    subcrates_folder = serato_folder / "SubCrates"
    crates = [
        Crate(name.stem) for name in subcrates_folder.glob("*.crate")
    ]
    all_crates.extend(crates)
    return all_crates

def make_crates():
    builder = Builder()
    lvl2_1 = Crate('lvl2_1')
    lvl2_1.add_track(Track.from_path("/Users/lukepurnell/test music/Russian Circles - Gnosis/Russian Circles - Gnosis - 06 Betrayal.mp3"))
    lvl2_2 = Crate('lvl2_2')
    lvl2_2.add_track(Track.from_path("/Users/lukepurnell/test music/Laker - Noise From The Ruliad/Laker - Noise From The Ruliad - 01 Escape.mp3"))
    lvl1_1 = Crate('lvl1_1', children=[lvl2_1, lvl2_2])
    lvl1_1.add_track(Track.from_path("/Users/lukepurnell/test music/Laker - Noise From The Ruliad/Laker - Noise From The Ruliad - 02 Points Break At Break Points.mp3"))
    root_crate = Crate('rootpyserato', children=[lvl1_1])
    builder.save(root_crate)


if __name__ == '__main__':
    make_crates()
    #    -root
    #        -lvl1_1
    #             - lvl2_1
    #             - lvl2_2

    #path = '/Users/lukepurnell/Music/_Serato_/Subcrates/all%%new serato crate%%nested%%new music.crate'
    #path = '/Users/lukepurnell/Music/_Serato_/Subcrates/all%%test.crate'
    #path = '/Users/lukepurnell/Music/_Serato_/Subcrates/all%%NOPLAYLIST.crate'
    #crate = builder.build_crates_from_filepath(Path(path))
    #print(crate)
    #lvl2_1 = Crate('lvl2_1')
    #print(f"crate: {lvl2_1}")
    #lvl2_1.add_song(Path("/Users/lukepurnell/nav_music/Russian Circles/Gnosis/01 Tupilak.wav"))
    #lvl1_1 = Crate('lvl1_1', children=[lvl2_1])
    #lvl1_1.add_song(Path("/Users/lukepurnell/nav_music/Laker/Noise From The Ruliad/00 Entropy Increasing.mp3"))
    #root_crate = Crate('root', children=[lvl1_1])
    #builder.save(root_crate)



    #child1_crate = Crate('test-child1')
    #child1_crate.add_song(Path("/Users/lukepurnell/Music/Russian Circles - Gnosis/Russian Circles - Gnosis - 03 Gnosis.mp3"))
    #child2_crate = Crate('test-child2', children=[child1_crate])
    #child2_crate.add_song(Path("/Users/lukepurnell/Music/Laker - Noise From The Ruliad/Laker - Noise From The Ruliad - 02 Points Break At Break Points.mp3"))
    #child3_crate = Crate('test-child3', children=[child1_crate])
    #child3_crate.add_song(Path("/Users/lukepurnell/Music/Laker - Noise From The Ruliad/Laker - Noise From The Ruliad - 07 To Compose To Decompose.mp3"))
    #root_crate = Crate('root', children=[child1_crate, child2_crate, child3_crate])
    #root_crate.add_song(Path("/Users/lukepurnell/Music/Justin Grounds - The Moon Pulled Us Here EP/Justin Grounds - The Moon Pulled Us Here EP - 02 Santa Margharita.mp3"))
    #crate.add_song(Path("/Users/lukepurnell/Music/serato_export_test/subbox_export/Cloudkicker/Solitude/05 - I’m Glad You Still Have a Sense of Humor….flac"))
    #crate.add_song(Path("/Users/lukepurnell/Music/beets/Arca/Arca/04 Urchin.mp3"))
    ##crate.add_song(Path("…"))
    #builder.save(root_crate)



from copy import deepcopy
from itertools import zip_longest
from typing import Optional
from typing_extensions import Self

from pyserato.model.track import Track
from pyserato.util import sanitize_filename, DuplicateTrackError


class Crate:
    def __init__(self, name: str, children: Optional[list[Self]] = None):
        self._children = children if children else []
        self.name = sanitize_filename(name)
        self._tracks: set[Track] = set()

    @property
    def children(self) -> list[Self]:
        return self._children

    @property
    def tracks(self) -> set[Track]:
        return self._tracks

    def add_track(self, track: Track) -> None:
        """
        Adds a unique Track to the Crate
        """
        if track in self._tracks:
            raise DuplicateTrackError(f"track {track} is already in the crate {self.name}")
        self._tracks.add(track)

    def __str__(self):
        return f"Crate<{self.name}>"

    def __repr__(self):
        return f"Crate<{self.name}>"

    def __add__(self, other) -> "Crate":
        assert self.name == other.name
        children = self._children + other._children
        children_copy = []
        for child in children:
            children_copy.append(deepcopy(child))
        new = Crate(self.name, children=children_copy)
        all_tracks = self._tracks | other._tracks
        for track in all_tracks:
            new.add_track(track)
        return new

    def __deepcopy__(self, memodict={}) -> "Crate":
        children_copy = []
        for child in self.children:
            children_copy.append(deepcopy(child))
        copy = Crate(self.name, children=children_copy)
        memodict[id(self)] = copy
        for track in self._tracks:
            copy.add_track(track)
        return copy

    def __eq__(self, other):
        result = self.name == other.name and self._tracks == other._tracks
        if self._children:
            # sort so order is deterministic
            children = sorted(self._children, key=lambda c: len(c.tracks))
            other_children = sorted(other._children, key=lambda c: len(c.tracks))
            for child, other_child in zip_longest(children, other_children):
                if child and other_child:
                    result &= child == other_child
                else:
                    result = False
        return result



from dataclasses import dataclass, field
from pathlib import Path
import logging
from typing import Optional

from pyserato.model.hot_cue import HotCue
from pyserato.model.hot_cue_type import HotCueType
from pyserato.model.tempo import Tempo

logger = logging.getLogger(__name__)


@dataclass
class Track:
    path: Path
    track_id: str = ""
    average_bpm: float = 0.0
    date_added: str = ""
    play_count: str = ""
    tonality: str = ""
    total_time: float = 0.0

    beatgrid: list[Tempo] = field(default_factory=list)
    hot_cues: list[HotCue] = field(default_factory=list)
    cue_loops: list[HotCue] = field(default_factory=list)

    @staticmethod
    def from_path(path: Path | str, user_root: Optional[Path] = None) -> "Track":
        """
        Adds a unique track path to the crate. Raises DuplicateTrackError if track path is already present in the Crate.
        :param track_path:
        :param user_root: Support adding an arbitrary root to the tracks.
        This is useful when run in a docker container and the path needs to refer to one on the host.
        :return:
        """
        if isinstance(path, str):
            path = Path(path)
        full_path = user_root / path if user_root else path
        # note the path on the system may not yet exist. This is acceptable. As long as the path of the track is present
        # on the host system where the Serato crates are located at the point of opening Serato, the tracks will be
        # found.
        # assert full_path.exists(), f"path of track does not exist {full_path}"
        resolved = full_path.expanduser().resolve()
        return Track(resolved)

    def add_beatgrid_marker(self, tempo: Tempo):
        self.beatgrid.append(tempo)

    def add_hot_cue(self, hot_cue: HotCue):
        assert len(self.hot_cues) < 8, "cannot have more than 8 hot cues on a track"
        assert len(self.cue_loops) < 4, "cannot have more than 4 loops on a track"
        at_index = hot_cue.index
        if hot_cue.type == HotCueType.LOOP:
            self.cue_loops.insert(at_index, hot_cue)
        else:
            self.hot_cues.insert(at_index, hot_cue)

    # TODO
    # def apply_beatgrid_offsets(self, offsets: list[Offset]):
    #     try:
    #         for hot_cue in self.hot_cues:
    #             hot_cue.offset = closest_offset(hot_cue.start, offsets)
    #             hot_cue.apply_offset()

    #         for loop in self.cue_loops:
    #             loop.offset = closest_offset(loop.start, offsets)
    #             loop.apply_offset()
    #     except ValueError as e:
    #         logger.error(f"Error: {e} | Track: {self.filename()}")

    def __eq__(self, other):
        return self.path == other.path

    def __hash__(self):
        return hash(self.path)



from abc import abstractmethod, ABC

from pyserato.model.track import Track


class BaseEncoder(ABC):

    @property
    @abstractmethod
    def tag_name(self) -> str:
        pass

    @property
    @abstractmethod
    def tag_version(self) -> bytes:
        pass

    @property
    @abstractmethod
    def markers_name(self) -> str:
        pass

    @abstractmethod
    def write(self, track: Track):
        pass



import base64
import struct
from io import BytesIO
from typing import List, Iterator

from mutagen.mp3 import MP3
from mutagen import id3

from pyserato.encoders.base_encoder import BaseEncoder
from pyserato.encoders.serato_tags import SERATO_MARKERS_V2
from pyserato.model.hot_cue import HotCue
from pyserato.model.hot_cue_type import HotCueType
from pyserato.model.track import Track
from pyserato.util import split_string


class V2Mp3Encoder(BaseEncoder):

    @property
    def fmt_version(self) -> str:
        return "BB"

    @property
    def tag_name(self) -> str:
        return SERATO_MARKERS_V2

    @property
    def tag_version(self) -> bytes:
        return b"\x01\x01"

    @property
    def markers_name(self) -> str:
        return "Serato Markers2"

    def write(self, track: Track):
        tagged_file = self._write(track, self._encode(track))
        tagged_file.save()

    def read_cues(self, track: Track) -> List[HotCue]:
        tags = MP3(track.path)
        tag_data = tags[self.tag_name]
        data = tag_data.data
        return list(self._decode(data))

    def _decode(self, data: bytes) -> Iterator[HotCue]:
        fp = BytesIO(data)
        assert struct.unpack(self.fmt_version, fp.read(2)) == (0x01, 0x01)
        payload = fp.read()
        data = b"".join(self._remove_null_padding(payload).split(b"\n"))
        data = self._pad_encoded_data(data)
        decoded = base64.b64decode(data)

        fp = BytesIO(decoded)
        assert struct.unpack(self.fmt_version, fp.read(2)) == (0x01, 0x01)

        while True:
            entry_name = self._get_entry_name(fp)  # NULL byte between name and length is already omitted
            if len(entry_name) == 0:
                break  # End of data

            struct_length = struct.unpack(">I", fp.read(4))[0]
            assert struct_length > 0  # normally this should not happen
            entry_data = fp.read(struct_length)

            match entry_name:
                case "COLOR":
                    # not yet implemented
                    continue
                case "CUE":
                    yield HotCue.from_bytes(entry_data, hotcue_type=HotCueType.CUE)
                case "LOOP":
                    yield HotCue.from_bytes(entry_data, hotcue_type=HotCueType.LOOP)
                case "BPMLOCK":
                    # not yet implemented
                    continue

    def _get_entry_count(self, buffer: BytesIO):
        return struct.unpack(">I", buffer.read(4))[0]

    def _remove_null_padding(self, payload: bytes):
        """
        Used when reading the data from the tags
        """
        return payload[: payload.index(b"\x00")]

    def _get_entry_name(self, fp) -> str:
        entry_name = b""
        for x in iter(lambda: fp.read(1), b""):
            if x == b"\00":
                return entry_name.decode("utf-8")

            entry_name += x

        return ""

    def _pad_encoded_data(self, data: bytes) -> bytes:
        """
        Used when reading the data from the tags
        """
        padding = b"A==" if len(data) % 4 == 1 else (b"=" * (-len(data) % 4))

        return data + padding

    def _write(self, track: Track, payload: bytes) -> MP3:
        mutagen_file = MP3(track.path)
        mutagen_file[self.tag_name] = id3.GEOB(
            encoding=0,
            mime="application/octet-stream",
            desc=self.markers_name,
            data=payload,
        )

        return mutagen_file

    def _encode(self, track: Track) -> bytes:
        payload = b""
        for cue in track.hot_cues:
            payload += cue.to_v2_bytes()
        for loop in track.cue_loops:
            payload += loop.to_v2_bytes()
        return self._pad(payload)

    def _pad(self, payload: bytes, entries_count: int | None = None):
        """
        Serato adds null padding at the end of the string.
        When the payload length is under 512 it pads until that number
        WHEN the payload is over 512 it pads until 1025

        Also, the payload is split at 72 characters before padding is applied
        """
        # Append the version for the non-encoded payload
        payload = self.tag_version + payload
        payload = self._remove_encoded_data_pad(base64.b64encode(payload))
        payload = self._pad_payload(split_string(payload))
        payload = self._enrich_payload(payload, entries_count)

        return payload

    @staticmethod
    def _remove_encoded_data_pad(data: bytes):
        """
        Used when after the base64 encode when writing data to the tags
        """
        return data.replace(b"=", b"A")

    @staticmethod
    def _pad_payload(payload: bytes):
        """
        Used when writing the data to the tags
        """
        length = len(payload)
        if length < 468:
            return payload.ljust(468, b"\x00")

        return payload.ljust(982, b"\x00") + b"\x00"

    def _enrich_payload(self, payload: bytes, entries_count: int | None = None):
        header = self.tag_version
        if entries_count is not None:
            header += struct.pack(">I", entries_count)

        return header + payload



from mutagen.id3 import ID3
from mutagen.mp3 import MP3

from pyserato.encoders.v2_mp3_encoder import V2Mp3Encoder
from pyserato.builder import Builder
from pyserato.model.track import Track
from pyserato.model.crate import Crate
from pyserato.model.hot_cue import HotCue
from pyserato.model.hot_cue_type import HotCueType

TAG_NAME = 'GEOB:Serato Markers2'
track_path = "/Users/lukepurnell/test music/Russian Circles - Gnosis/Russian Circles - Gnosis - 06 Betrayal.mp3"
TAG_NAME_V1 = 'GEOB:Serato Overview'
TAG_NAME_V1_2 = 'GEOB:Serato Markers_'
TAG_NAME_V1_3 = 'GEOB:Serato Analysis'


def clear_tags():
    track = MP3(track_path)
    tag_names = [TAG_NAME, TAG_NAME_V1, TAG_NAME_V1_2, TAG_NAME_V1_3]
    #tag_names = [TAG_NAME]
    for tn in tag_names:
        try:
            data = track.pop(tn)
        except Exception:
            pass
        else:
            print(data)
    track.save()

def read_tags():
    mp3_encoder = V2Mp3Encoder()
    track = Track.from_path(track_path)
    cues = mp3_encoder.read_cues(track)
    for cue in cues:
        print(cue)

def get_geob_tabs():
    tags = ID3(track_path)
    for frame in tags.getall("GEOB"):
        print(frame)

def write_tags():
    mp3_encoder = V2Mp3Encoder()
    builder = Builder(encoder=mp3_encoder)
    crate = Crate('foopython')
    track = Track.from_path(track_path)
    crate.add_track(track)
    track.add_hot_cue(HotCue(name='cue1', type=HotCueType.CUE, start=50, index=1))
    track.add_hot_cue(HotCue(name='looppython1', type=HotCueType.LOOP, start=1900, end=3600, index=1))
    track.add_hot_cue(HotCue(name='looppython2', type=HotCueType.LOOP, start=4900, end=8600, index=2))
    builder.save(crate, overwrite=True)
    # or
    #mp3_encoder.write(track)

if __name__ == '__main__':
    #write_tags2()
    #clear_tags()
    write_tags()
    #get_geob_tabs()
    read_tags()



from pyserato.util import serato_encode, serato_decode


def test_serato_encode_decode():
    test_s = "/Users/lukepurnell/Music/beets/Arca/Arca/10 Desafío.mp3"
    assert serato_decode(serato_encode(test_s)) == test_s



import struct


def encode(data):
    """Encode 3 byte plain text into 4 byte Serato binary format."""
    a, b, c = struct.unpack("BBB", data)
    z = c & 0x7F
    y = ((c >> 7) | (b << 1)) & 0x7F
    x = ((b >> 6) | (a << 2)) & 0x7F
    w = a >> 5
    return bytes(bytearray([w, x, y, z]))


def decode(data):
    """Decode 4 byte Serato binary format into 3 byte plain text."""
    w, x, y, z = struct.unpack("BBBB", data)
    c = (z & 0x7F) | ((y & 0x01) << 7)
    b = ((y & 0x7F) >> 1) | ((x & 0x03) << 6)
    a = ((x & 0x7F) >> 2) | ((w & 0x07) << 5)
    return struct.pack("BBB", a, b, c)



from pyserato.encoders.base_encoder import BaseEncoder
from pyserato.encoders.serato_tags import SERATO_MARKERS_V1


class V1Mp3Encoder(BaseEncoder):
    STRUCT_LENGTH = 0x16

    @property
    def tag_name(self) -> str:
        return SERATO_MARKERS_V1

    @property
    def tag_version(self) -> bytes:
        return b"\x02\x05"

    @property
    def markers_name(self) -> str:
        return "Serato Markers_"



from pathlib import Path

from mutagen.mp3 import MP3

SERATO_MARKERS_V2 = "GEOB:Serato Markers2"
SERATO_OVERVIEW = "GEOB:Serato Overview"
SERATO_MARKERS_V1 = "GEOB:Serato Markers_"
SERATO_ANALYSIS = "GEOB:Serato Analysis"


def clear_all_tags(track_path: Path, audio_encoding="mp3"):
    assert audio_encoding == "mp3", "only mp3 supported"
    track = MP3(track_path)
    for tag in [SERATO_MARKERS_V2, SERATO_OVERVIEW, SERATO_MARKERS_V1, SERATO_ANALYSIS]:
        try:
            track.pop(tag)
        except Exception:
            pass
    track.save()



import re

INVALID_CHARACTERS_REGEX = re.compile(r"[^A-Za-z0-9_ ]", re.IGNORECASE)


def split_string(string: bytes, after: int = 72, delimiter: bytes = b"\n"):
    pieces = []
    while len(string) > 0:
        pieces.append(string[:after])
        string = string[after:]

    return delimiter.join(pieces)


def serato_decode(s: bytes) -> str:
    """
    Decode a string that's been encoded in to bytes Serato style.
    This is a Python implementation of Java's bytes to string utf16 which Serato appears to use from looking at:
    https://github.com/markusschmitz53/serato-itch-sync
    :param s:
    :return:
    """
    result = ""
    while s:
        chunk = s[:2]
        chunk = chunk[::-1]
        result += chunk.decode("utf16")
        s = s[2:]
    return result


def serato_encode(s: str) -> bytes:
    """
    Encode a string Serato style.
    This is a Python implementation of Java's 'writeChars' which Serato appears to use from looking at:
    https://github.com/markusschmitz53/serato-itch-sync
    :param s:
    :return:
    """
    result = []
    for c in s:
        result.append(int.from_bytes(c.encode("utf16"), byteorder="big") >> 0 & 255)
        result.append(int.from_bytes(c.encode("utf16"), byteorder="big") >> 8 & 255)
    return bytes(result)


def sanitize_filename(filename: str) -> str:
    return re.sub(INVALID_CHARACTERS_REGEX, "-", filename)


class DuplicateTrackError(Exception):
    """
    Raised if a track is added to a crate for which the track's path resolves to an existing track path already in the
    crate.
    """



from enum import Enum


class SeratoColor(str, Enum):
    """
    https://github.com/Holzhaus/serato-tags/blob/main/docs/colors.md#serato-dj-pro
    For Serato DJ Pro, saved and displayed colors are different. This map is the saved colors.
    """

    RED = "CC0000"
    ORANGE = "CC4400"
    AMBER = "CC8800"
    YELLOW = "CCCC00"
    LIME_GREEN = "88CC00"
    GREEN_YELLOW = "44CC00"
    GREEN = "00CC00"
    MINT_GREEN = "00CC44"
    TEAL_GREEN = "00CC88"
    TEAL = "00CCCC"
    SKY_BLUE = "0088CC"
    BLUE = "0044CC"
    DARK_BLUE = "0000CC"
    PURPLE = "4400CC"
    VIOLET = "8800CC"
    MAGENTA = "CC00CC"
    PINK = "CC0088"
    RED_PINK = "CC0044"
